#ifdef MNH_OPENACC
!MNH_LIC Copyright 2013-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
! Modifications:
!  P. Wautelet 10/07/2019: bugfix: MNH_REL_ZT3D_N0: access outside of array was possible
!-----------------------------------------------------------------
MODULE MODE_MNH_ZWORK

  use mode_msg

  IMPLICIT NONE

  INTEGER, SAVE :: IIB,IJB,IKB  ! Begining useful area in x,y,z directions
  INTEGER, SAVE :: IIE,IJE,IKE  ! End useful area in x,y,z directions

  !
  INTEGER,SAVE                    :: IJS,IJN, IIW,IIA
  !
  INTEGER, SAVE                   :: IIU,IJU,IKU,IIJKU
  LOGICAL, SAVE                   :: GWEST , GEAST
  LOGICAL, SAVE                   :: GSOUTH , GNORTH

  LOGICAL, SAVE                   :: GFIRST_CALL_MNH_ALLOC_ZWORK = .TRUE.
  !
  REAL, SAVE, ALLOCATABLE , DIMENSION(:,:)   :: ZPSRC_HALO2_WEST
  REAL, SAVE, ALLOCATABLE , DIMENSION(:,:)   :: ZPSRC_HALO2_SOUTH

  REAL, SAVE, ALLOCATABLE , DIMENSION(:,:,:) :: ZUNIT3D

  INTEGER, parameter                        :: JPMAX_T3D = 100
  INTEGER , ALLOCATABLE, DIMENSION (:)      :: NT3D_POOL
  INTEGER                                   :: NT3D_TOP , NT3D_TOP_MAX = 0
  INTEGER                                   :: NT3D_TOP_CURRENT(JPMAX_T3D+1) ,  NT3D_TOP_CURRENT_INDEX = 0
  !REAL    , ALLOCATABLE, DIMENSION(:,:,:,:) , TARGET :: ZT3D_A1,ZT3D_A2,ZT3D_A3,ZT3D_A4
  !REAL    , POINTER    , DIMENSION(:,:,:,:)          :: ZT3D
  REAL,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:,:,:,:)          :: ZT3D

  REAL,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:)                :: ZT1D_OSIZE
  REAL,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:,:)              :: ZT2D_OSIZE
  REAL,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:,:,:)            :: ZT3D_OSIZE
  REAL,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:,:,:,:)          :: ZT4D_OSIZE
  

  TYPE TMODEL
      REAL    , POINTER, DIMENSION(:,:,:,:) :: X
  END TYPE TMODEL

  TYPE(TMODEL) , DIMENSION(10) :: MODEL

  INTEGER, parameter                        :: JPMAX_T3D_G = 12
  INTEGER , ALLOCATABLE, DIMENSION (:)      :: NT3D_POOL_G
  INTEGER                                   :: NT3D_TOP_G , NT3D_TOP_G_MAX = 0  
  LOGICAL,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:,:,:,:)       :: GT3D
  LOGICAL,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:,:)           :: GT2D_OSIZE

  INTEGER, parameter                        :: JPMAX_T3D_I = 4
  INTEGER , ALLOCATABLE, DIMENSION (:)      :: NT3D_POOL_I
  INTEGER                                   :: NT3D_TOP_I , NT3D_TOP_I_MAX = 0  
  INTEGER,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:,:,:,:)       :: IT3D

  INTEGER, parameter                        :: JPMAX_T1D_R = 60
  INTEGER , ALLOCATABLE, DIMENSION (:)      :: NT1D_POOL_R
  INTEGER                                   :: NT1D_TOP_R , NT1D_TOP_R_MAX = 0  
  REAL,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:,:)       :: ZT1D 
  
  INTEGER, parameter                        :: JPMAX_T1D_I = 30
  INTEGER , ALLOCATABLE, DIMENSION (:)      :: NT1D_POOL_I
  INTEGER                                   :: NT1D_TOP_I , NT1D_TOP_I_MAX = 0  
  INTEGER,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:,:)       :: IT1D
  INTEGER,SAVE    , ALLOCATABLE, TARGET , DIMENSION(:)         :: IT1D_OSIZE
  
CONTAINS

  SUBROUTINE MNH_ALLOC_ZWORK(IMODEL)

    USE MODE_TOOLS_ll, ONLY   : GET_DIM_EXT_ll, GET_INDICE_ll, LWEST_ll,LEAST_ll, LSOUTH_ll, LNORTH_ll

    USE MODD_PARAMETERS, ONLY : JPVEXT, XUNDEF
    USE MODD_DIM_n,      ONLY : NKMAX

    IMPLICIT NONE

    INTEGER :: IMODEL

    INTEGER :: JI

    IF (GFIRST_CALL_MNH_ALLOC_ZWORK) THEN
       GFIRST_CALL_MNH_ALLOC_ZWORK = .FALSE.
       !
       ! Array dim
       !
       CALL GET_DIM_EXT_ll('B',IIU,IJU)     
       IKU=NKMAX + 2* JPVEXT
       IIJKU = IIU*IJU*IKU
       !
       ! Computation bound
       !
       CALL GET_INDICE_ll(IIB,IJB,IIE,IJE)
       IJS=IJB
       IJN=IJE
       IIW=IIB
       IIA=IIE
       IKB = 1     + JPVEXT
       IKE = NKMAX + JPVEXT
       !
       ! Lateral boundary
       !
       GWEST = LWEST_ll()
       GEAST = LEAST_ll()

       GSOUTH=LSOUTH_ll()
       GNORTH=LNORTH_ll()

       !
       ! Work array
       ! 
       ALLOCATE (ZPSRC_HALO2_WEST(IJU,IKU)) 
       ALLOCATE (ZPSRC_HALO2_SOUTH(IIU,IKU)) 
       !$acc enter data create(ZPSRC_HALO2_WEST,ZPSRC_HALO2_SOUTH)

       ALLOCATE (ZUNIT3D(IIU,IJU,IKU))
       !$acc enter data create(ZUNIT3D)

!----- Real pool
       
       !ALLOCATE (ZT3D_A1(IIU,IJU,IKU,JPMAX_T3D))
       !MODEL(1)%X => ZT3D_A1
       !ZT3D => MODEL(1)%X
       ALLOCATE (ZT3D(IIU,IJU,IKU,JPMAX_T3D))
       !$acc enter data create(ZT3D)

       ALLOCATE (ZT1D_OSIZE(0))
       ALLOCATE (ZT2D_OSIZE(IIU,0))
       ALLOCATE (ZT3D_OSIZE(IIU,IJU,0))
       ALLOCATE (ZT4D_OSIZE(IIU,IJU,IKU,0))
       !$acc enter data create(ZT1D_OSIZE,ZT2D_OSIZE,ZT3D_OSIZE,ZT4D_OSIZE)      

       ALLOCATE (NT3D_POOL(JPMAX_T3D))
       NT3D_TOP = 0
       DO JI = 1, JPMAX_T3D
          NT3D_POOL(JI) = JI
       END DO

!------ Logical pool
       
       ALLOCATE (GT3D(IIU,IJU,IKU,JPMAX_T3D_G))
       !$acc enter data create(GT3D)

       ALLOCATE (GT2D_OSIZE(IIU,0))
       !$acc enter data create(GT2D_OSIZE)

       ALLOCATE (NT3D_POOL_G(JPMAX_T3D_G))
       NT3D_TOP_G = 0
       DO JI = 1, JPMAX_T3D_G
          NT3D_POOL_G(JI) = JI
       END DO

!------ Integer pool
       
       ALLOCATE (IT3D(IIU,IJU,IKU,JPMAX_T3D_I))
       !$acc enter data create(IT3D)

       ALLOCATE (NT3D_POOL_I(JPMAX_T3D_I))
       NT3D_TOP_I = 0
       DO JI = 1, JPMAX_T3D_I
          NT3D_POOL_I(JI) = JI
       END DO

!------ Real 1D pool 
       
       ALLOCATE (ZT1D(IIU*IJU*IKU,JPMAX_T1D_R))
       !$acc enter data create(ZT1D)

       ALLOCATE (NT1D_POOL_R(JPMAX_T1D_R))
       NT1D_TOP_R = 0
       DO JI = 1, JPMAX_T1D_R
          NT1D_POOL_R(JI) = JI
       END DO

!------ Integer 1D pool
       
       ALLOCATE (IT1D(IIU*IJU*IKU,JPMAX_T1D_I))
       ALLOCATE (IT1D_OSIZE(0))
       !$acc enter data create(IT1D,IT1D_OSIZE)

       ALLOCATE (NT1D_POOL_I(JPMAX_T1D_I))
       NT1D_TOP_I = 0
       DO JI = 1, JPMAX_T1D_I
          NT1D_POOL_I(JI) = JI
       END DO       
       
!------ Default values
       
       !$acc kernels 

       ZPSRC_HALO2_WEST = XUNDEF
       ZPSRC_HALO2_SOUTH = XUNDEF

       ZUNIT3D = 1.0

       ZT3D  =  XUNDEF
       ZT1D  =  XUNDEF

       IT3D  = 0.0
       IT1D  = 0.0

       GT3D  = .FALSE.

       !$acc end kernels

       !$acc update host (ZPSRC_HALO2_WEST,ZPSRC_HALO2_SOUTH)
       !$acc update host (ZUNIT3D)
       !$acc update host (ZT3D,ZT1D)

    END IF

  END SUBROUTINE MNH_ALLOC_ZWORK

  SUBROUTINE MNH_GET_ZT3D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF (NT3D_TOP == JPMAX_T3D ) THEN
      WRITE( *, '( " MNH_GET_ZT3D_N0: NT3D_TOP too big (increaze JPMAX_T3D) , NT3D_TOP=",I4 )' ) NT3D_TOP
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_ZT3D_N0', 'NT3D_TOP too big (increaze JPMAX_T3D)' )
    ELSE
       NT3D_TOP = NT3D_TOP + 1
       KTEMP = NT3D_POOL(NT3D_TOP)
       NT3D_POOL(NT3D_TOP) = - KTEMP
       IF (  NT3D_TOP > NT3D_TOP_MAX ) THEN
          NT3D_TOP_MAX = NT3D_TOP
          !WRITE( *, '( " MNH_GET_ZT3D: NT3D_TOP_MAX=",I4," KTEMP=",I4 )' ) NT3D_TOP_MAX,KTEMP
       END IF
    ENDIF
    !WRITE( *, '( "MNH_GET_ZT3D: reserving ZT3D (",I4,")" )' ) KTEMP

  END SUBROUTINE MNH_GET_ZT3D_N0

  SUBROUTINE MNH_GET_ZT3D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18

                          CALL MNH_GET_ZT3D_N0(KTEMP1)
    IF (PRESENT(KTEMP2))  CALL MNH_GET_ZT3D_N0(KTEMP2)
    IF (PRESENT(KTEMP3))  CALL MNH_GET_ZT3D_N0(KTEMP3)
    IF (PRESENT(KTEMP4))  CALL MNH_GET_ZT3D_N0(KTEMP4)
    IF (PRESENT(KTEMP5))  CALL MNH_GET_ZT3D_N0(KTEMP5)
    IF (PRESENT(KTEMP6))  CALL MNH_GET_ZT3D_N0(KTEMP6)
    IF (PRESENT(KTEMP7))  CALL MNH_GET_ZT3D_N0(KTEMP7)
    IF (PRESENT(KTEMP8))  CALL MNH_GET_ZT3D_N0(KTEMP8)
    IF (PRESENT(KTEMP9))  CALL MNH_GET_ZT3D_N0(KTEMP9)
    IF (PRESENT(KTEMP10)) CALL MNH_GET_ZT3D_N0(KTEMP10)
    IF (PRESENT(KTEMP11)) CALL MNH_GET_ZT3D_N0(KTEMP11)
    IF (PRESENT(KTEMP12)) CALL MNH_GET_ZT3D_N0(KTEMP12)
    IF (PRESENT(KTEMP13)) CALL MNH_GET_ZT3D_N0(KTEMP13)
    IF (PRESENT(KTEMP14)) CALL MNH_GET_ZT3D_N0(KTEMP14)
    IF (PRESENT(KTEMP15)) CALL MNH_GET_ZT3D_N0(KTEMP15)
    IF (PRESENT(KTEMP16)) CALL MNH_GET_ZT3D_N0(KTEMP16)
    IF (PRESENT(KTEMP17)) CALL MNH_GET_ZT3D_N0(KTEMP17)
    IF (PRESENT(KTEMP18)) CALL MNH_GET_ZT3D_N0(KTEMP18)

  END SUBROUTINE MNH_GET_ZT3D

  SUBROUTINE MNH_GET_ZT4D(KSIZE,KBEG,KEND)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: KSIZE
    INTEGER, INTENT(OUT) :: KBEG, KEND

    INTEGER :: JI

    IF (NT3D_TOP + KSIZE > JPMAX_T3D ) THEN
      WRITE( *, '( " MNH_GET_ZT4D: NT3D_TOP+KSIZE too big (increaze JPMAX_T3D) , NT3D_TOP=",I4 )' ) NT3D_TOP+KSIZE
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_ZT4D', 'NT3D_TOP+ KSIZE too big (increaze JPMAX_T3D)' )
    ELSE
       KBEG = NT3D_TOP + 1
       KEND = NT3D_TOP + KSIZE
       NT3D_TOP = NT3D_TOP + KSIZE
       DO JI = KBEG, KEND
         IF (NT3D_POOL(JI) <= 0 ) THEN           
           call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_ZT4D', 'trying to use area already reserved' )
         END IF
         NT3D_POOL(JI) = -JI
       END DO
       IF ( NT3D_TOP > NT3D_TOP_MAX ) THEN
          NT3D_TOP_MAX = NT3D_TOP
          WRITE( *, '( " MNH_GET_ZT4D: NT3D_TOP_MAX=",I4," KBEG=",I4," KEND=",I4 )' ) NT3D_TOP_MAX,KBEG,KEND
       END IF
    ENDIF
    !WRITE( *, '( "MNH_GET_ZT4D: reserving ZT3D (",I4,I4,")" )' ) KBEG,KEND

  END SUBROUTINE MNH_GET_ZT4D

  SUBROUTINE MNH_REL_ZT3D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF ( KTEMP .EQ. 0 ) THEN
       ! Special case Zero size array do nothing
       RETURN
    ELSE
    
    IF ( ( NT3D_TOP > JPMAX_T3D ) .OR. ( NT3D_TOP < 1 ) ) THEN
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_ZT3D_N0', 'invalid value for NT3D_TOP' )
    ELSE
      NT3D_POOL(KTEMP) = KTEMP
      IF (KTEMP == NT3D_TOP) THEN
        NT3D_TOP = NT3D_TOP - 1
     ELSE
        WRITE( *, '( "MNH_REL_ZT3D: releasing ZT3D (",2I8,")" )' ) KTEMP, NT3D_TOP
        call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_ZT3D_N0', 'invalid value for KTEMP <> NT3D_TOP' )
      END IF
    ENDIF
    !WRITE( *, '( "MNH_REL_ZT3D: releasing ZT3D (",I4,")" )' ) KTEMP
    ENDIF

  END SUBROUTINE MNH_REL_ZT3D_N0

  SUBROUTINE MNH_REL_ZT3D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18, &
                          KTEMP19,KTEMP20)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18
    INTEGER,OPTIONAL :: KTEMP19,KTEMP20

    IF (PRESENT(KTEMP20)) CALL MNH_REL_ZT3D_N0(KTEMP20)
    IF (PRESENT(KTEMP19)) CALL MNH_REL_ZT3D_N0(KTEMP19)
    IF (PRESENT(KTEMP18)) CALL MNH_REL_ZT3D_N0(KTEMP18)
    IF (PRESENT(KTEMP17)) CALL MNH_REL_ZT3D_N0(KTEMP17)
    IF (PRESENT(KTEMP16)) CALL MNH_REL_ZT3D_N0(KTEMP16)
    IF (PRESENT(KTEMP15)) CALL MNH_REL_ZT3D_N0(KTEMP15)
    IF (PRESENT(KTEMP14)) CALL MNH_REL_ZT3D_N0(KTEMP14)
    IF (PRESENT(KTEMP13)) CALL MNH_REL_ZT3D_N0(KTEMP13)
    IF (PRESENT(KTEMP12)) CALL MNH_REL_ZT3D_N0(KTEMP12)
    IF (PRESENT(KTEMP11)) CALL MNH_REL_ZT3D_N0(KTEMP11)
    IF (PRESENT(KTEMP10)) CALL MNH_REL_ZT3D_N0(KTEMP10)
    IF (PRESENT(KTEMP9))  CALL MNH_REL_ZT3D_N0(KTEMP9)
    IF (PRESENT(KTEMP8))  CALL MNH_REL_ZT3D_N0(KTEMP8)
    IF (PRESENT(KTEMP7))  CALL MNH_REL_ZT3D_N0(KTEMP7)
    IF (PRESENT(KTEMP6))  CALL MNH_REL_ZT3D_N0(KTEMP6)
    IF (PRESENT(KTEMP5))  CALL MNH_REL_ZT3D_N0(KTEMP5)
    IF (PRESENT(KTEMP4))  CALL MNH_REL_ZT3D_N0(KTEMP4)
    IF (PRESENT(KTEMP3))  CALL MNH_REL_ZT3D_N0(KTEMP3)
    IF (PRESENT(KTEMP2))  CALL MNH_REL_ZT3D_N0(KTEMP2)
                          CALL MNH_REL_ZT3D_N0(KTEMP1)
    
  END SUBROUTINE MNH_REL_ZT3D

  SUBROUTINE MNH_REL_ZT4D(KSIZE,KBEG)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: KSIZE
    INTEGER, INTENT(IN) :: KBEG

    character(len=16) :: ytxt1, ytxt2
    INTEGER :: JI

    IF ( KSIZE .EQ. 0 ) THEN
       ! special case of O zero 4D array => ZT4D_OSIZE
       RETURN
    END IF
    IF ( KBEG + KSIZE -1 /= NT3D_TOP ) THEN
      write( ytxt1, '( I4, "-", I4 )' ) kbeg, kbeg + ksize - 1
      write( ytxt2, '( I4 )' )          NT3D_TOP
      call Print_msg( NVERB_ERROR, 'GEN', 'MNH_REL_ZT4D', 'trying to free area (' // trim( ytxt1 ) // &
                                                          ') not at the end of range (' // trim( ytxt2 ) // ')' )
    END IF
    IF ( ( KBEG + KSIZE > JPMAX_T3D ) .OR. ( KBEG < 1 ) ) THEN
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_ZT4D', 'out of range' )
    END IF

    DO JI = KBEG, KBEG+KSIZE-1
       IF (NT3D_POOL(JI) /= - JI ) THEN        
          call Print_msg( NVERB_ERROR, 'GEN', 'MNH_REL_ZT4D', 'trying to free area not reserved' )
       ELSE
          NT3D_POOL(JI) = JI
      END IF
    END DO
    NT3D_TOP = NT3D_TOP - KSIZE
    !WRITE( *, '( "MNH_REL_ZT4D: releasing ZT3D (",I4,I4,")" )' ) KBEG,KBEG+KSIZE-1

  END SUBROUTINE MNH_REL_ZT4D

  FUNCTION MNH_ALLOCATE_ZT3D(PTAB,KI,KJ,KK) RESULT (KINDEX)
    
    REAL, POINTER, CONTIGUOUS , DIMENSION(:,:,:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI,KJ,KK
    INTEGER                 :: KINDEX

    !local

    IF ( (KI .EQ. IIU) .AND. (KJ .EQ. IJU) ) THEN
       CALL MNH_GET_ZT3D_N0(KINDEX)
       IF       (KK .LE. IKU)  THEN           
          PTAB => ZT3D(:,:,1:KK,KINDEX)
       ELSE IF (KK .EQ.0 )  THEN
          PTAB => ZT3D_OSIZE
       ELSE
          call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_ZT3D', ' Size mismatsh ' )
       END IF
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_ZT3D', ' Size mismatsh ' )
    END IF
       
  END FUNCTION MNH_ALLOCATE_ZT3D

    FUNCTION MNH_ALLOCATE_ZT4D(PTAB,KI,KJ,KK,KL) RESULT (KINDEX_BEG)
    
    REAL, POINTER, CONTIGUOUS , DIMENSION(:,:,:,:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI,KJ,KK,KL
    INTEGER                 :: KINDEX_BEG

    !local

    INTEGER                 :: KINDEX_END

    IF ( (KI .EQ. IIU) .AND. (KJ .EQ. IJU) .AND. (KK .EQ. IKU) ) THEN
       IF ( KL .GE. 1 ) THEN
          CALL MNH_GET_ZT4D(KL,KINDEX_BEG,KINDEX_END)
          PTAB => ZT3D(:,:,:,KINDEX_BEG:KINDEX_END)
       ELSE
          PTAB => ZT4D_OSIZE
          KINDEX_BEG = 0
       END IF
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_ZT4D', ' Size mismatsh ' )
    END IF
       
  END FUNCTION MNH_ALLOCATE_ZT4D

  FUNCTION MNH_ALLOCATE_ZT2D(PTAB,KI,KJ) RESULT (KINDEX)
    
    REAL, POINTER, CONTIGUOUS , DIMENSION(:,:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI,KJ
    INTEGER                 :: KINDEX

    !local

    IF (KI .EQ. IIU) THEN
       CALL MNH_GET_ZT3D_N0(KINDEX)
       IF (KJ .EQ. IJU) THEN 
          PTAB => ZT3D(:,:,1,KINDEX)
       ELSE IF (KJ .EQ. 0)  THEN
          PTAB => ZT2D_OSIZE
       ELSE
          call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_ZT2D', ' Size mismatsh ' )
       END IF
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_ZT2D', ' Size mismatsh ' )
    END IF
       
  END FUNCTION MNH_ALLOCATE_ZT2D

  FUNCTION MNH_ALLOCATE_ZT3DP(PTAB,KI,KJ,KKB,KKE) RESULT (KINDEX)
    
    REAL, POINTER, CONTIGUOUS , DIMENSION(:,:,:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI,KJ,KKB,KKE
    INTEGER                 :: KINDEX

    !local

    IF ( (KI .EQ. IIU) .AND. (KJ .EQ. IJU) .AND. (KKB .LE. IKU) .AND. (KKE  .LE. IKU) ) THEN 
       CALL MNH_GET_ZT3D_N0(KINDEX)
       PTAB(1:,1:,KKB:) => ZT3D(:,:,KKB:KKE,KINDEX)
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_ZT3DP', ' Size mismatsh ' )
    END IF
       
  END FUNCTION MNH_ALLOCATE_ZT3DP

!-------- Logical Pool Managment
  
    SUBROUTINE MNH_GET_GT3D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF (NT3D_TOP_G == JPMAX_T3D_G ) THEN
      WRITE( *, '( " MNH_GET_GT3D_N0: NT3D_TOP_G too big (increaze JPMAX_T3D_G) , NT3D_TOP_G=",I4 )' ) NT3D_TOP_G
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_GT3D_N0', 'NT3D_TOP_G too big (increaze JPMAX_T3D_G)' )
    ELSE
       NT3D_TOP_G = NT3D_TOP_G + 1
       KTEMP = NT3D_POOL_G(NT3D_TOP_G)
       NT3D_POOL_G(NT3D_TOP_G) = - KTEMP
       IF (  NT3D_TOP_G > NT3D_TOP_G_MAX ) THEN
          NT3D_TOP_G_MAX = NT3D_TOP_G
          !WRITE( *, '( " MNH_GET_GT3D: NT3D_TOP_G_MAX=",I4," KTEMP=",I4 )' ) NT3D_TOP_G_MAX,KTEMP
       END IF
    ENDIF
    !WRITE( *, '( "MNH_GET_GT3D: reserving GT3D (",I4,")" )' ) KTEMP

  END SUBROUTINE MNH_GET_GT3D_N0

  SUBROUTINE MNH_GET_GT3D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18

                          CALL MNH_GET_GT3D_N0(KTEMP1)
    IF (PRESENT(KTEMP2))  CALL MNH_GET_GT3D_N0(KTEMP2)
    IF (PRESENT(KTEMP3))  CALL MNH_GET_GT3D_N0(KTEMP3)
    IF (PRESENT(KTEMP4))  CALL MNH_GET_GT3D_N0(KTEMP4)
    IF (PRESENT(KTEMP5))  CALL MNH_GET_GT3D_N0(KTEMP5)
    IF (PRESENT(KTEMP6))  CALL MNH_GET_GT3D_N0(KTEMP6)
    IF (PRESENT(KTEMP7))  CALL MNH_GET_GT3D_N0(KTEMP7)
    IF (PRESENT(KTEMP8))  CALL MNH_GET_GT3D_N0(KTEMP8)
    IF (PRESENT(KTEMP9))  CALL MNH_GET_GT3D_N0(KTEMP9)
    IF (PRESENT(KTEMP10)) CALL MNH_GET_GT3D_N0(KTEMP10)
    IF (PRESENT(KTEMP11)) CALL MNH_GET_GT3D_N0(KTEMP11)
    IF (PRESENT(KTEMP12)) CALL MNH_GET_GT3D_N0(KTEMP12)
    IF (PRESENT(KTEMP13)) CALL MNH_GET_GT3D_N0(KTEMP13)
    IF (PRESENT(KTEMP14)) CALL MNH_GET_GT3D_N0(KTEMP14)
    IF (PRESENT(KTEMP15)) CALL MNH_GET_GT3D_N0(KTEMP15)
    IF (PRESENT(KTEMP16)) CALL MNH_GET_GT3D_N0(KTEMP16)
    IF (PRESENT(KTEMP17)) CALL MNH_GET_GT3D_N0(KTEMP17)
    IF (PRESENT(KTEMP18)) CALL MNH_GET_GT3D_N0(KTEMP18)

  END SUBROUTINE MNH_GET_GT3D

    SUBROUTINE MNH_REL_GT3D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF ( ( NT3D_TOP_G > JPMAX_T3D_G ) .OR. ( NT3D_TOP_G < 1 ) ) THEN
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_GT3D_N0', 'invalid value for NT3D_TOP_G' )
    ELSE
      NT3D_POOL_G(KTEMP) = KTEMP
      IF (KTEMP == NT3D_TOP_G) THEN
        NT3D_TOP_G = NT3D_TOP_G - 1
     ELSE
        WRITE( *, '( "MNH_REL_GT3D: releasing GT3D (",2I8,")" )' ) KTEMP, NT3D_TOP_G
        call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_GT3D_N0', 'invalid value for KTEMP <> NT3D_TOP_G' )
      END IF
    ENDIF
    !WRITE( *, '( "MNH_REL_GT3D: releasing GT3D (",I4,")" )' ) KTEMP

  END SUBROUTINE MNH_REL_GT3D_N0

  SUBROUTINE MNH_REL_GT3D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18, &
                          KTEMP19,KTEMP20)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18
    INTEGER,OPTIONAL :: KTEMP19,KTEMP20

    IF (PRESENT(KTEMP20)) CALL MNH_REL_GT3D_N0(KTEMP20)
    IF (PRESENT(KTEMP19)) CALL MNH_REL_GT3D_N0(KTEMP19)
    IF (PRESENT(KTEMP18)) CALL MNH_REL_GT3D_N0(KTEMP18)
    IF (PRESENT(KTEMP17)) CALL MNH_REL_GT3D_N0(KTEMP17)
    IF (PRESENT(KTEMP16)) CALL MNH_REL_GT3D_N0(KTEMP16)
    IF (PRESENT(KTEMP15)) CALL MNH_REL_GT3D_N0(KTEMP15)
    IF (PRESENT(KTEMP14)) CALL MNH_REL_GT3D_N0(KTEMP14)
    IF (PRESENT(KTEMP13)) CALL MNH_REL_GT3D_N0(KTEMP13)
    IF (PRESENT(KTEMP12)) CALL MNH_REL_GT3D_N0(KTEMP12)
    IF (PRESENT(KTEMP11)) CALL MNH_REL_GT3D_N0(KTEMP11)
    IF (PRESENT(KTEMP10)) CALL MNH_REL_GT3D_N0(KTEMP10)
    IF (PRESENT(KTEMP9))  CALL MNH_REL_GT3D_N0(KTEMP9)
    IF (PRESENT(KTEMP8))  CALL MNH_REL_GT3D_N0(KTEMP8)
    IF (PRESENT(KTEMP7))  CALL MNH_REL_GT3D_N0(KTEMP7)
    IF (PRESENT(KTEMP6))  CALL MNH_REL_GT3D_N0(KTEMP6)
    IF (PRESENT(KTEMP5))  CALL MNH_REL_GT3D_N0(KTEMP5)
    IF (PRESENT(KTEMP4))  CALL MNH_REL_GT3D_N0(KTEMP4)
    IF (PRESENT(KTEMP3))  CALL MNH_REL_GT3D_N0(KTEMP3)
    IF (PRESENT(KTEMP2))  CALL MNH_REL_GT3D_N0(KTEMP2)
    CALL MNH_REL_GT3D_N0(KTEMP1)
    
  END SUBROUTINE MNH_REL_GT3D

    FUNCTION MNH_ALLOCATE_GT3D(PTAB,KI,KJ,KK) RESULT (KINDEX)
    
    LOGICAL, POINTER, CONTIGUOUS , DIMENSION(:,:,:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI,KJ,KK
    INTEGER                 :: KINDEX

    !local

    IF ( (KI .EQ. IIU) .AND. (KJ .EQ. IJU) .AND. (KK .EQ. IKU) ) THEN 
       CALL MNH_GET_GT3D_N0(KINDEX)
       PTAB => GT3D(:,:,:,KINDEX)
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_GT3D', ' Size mismatsh ' )
    END IF
       
  END FUNCTION MNH_ALLOCATE_GT3D

  FUNCTION MNH_ALLOCATE_GT2D(PTAB,KI,KJ) RESULT (KINDEX)
    
    LOGICAL, POINTER, CONTIGUOUS , DIMENSION(:,:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI,KJ
    INTEGER                 :: KINDEX

    !local

    IF (KI .EQ. IIU) THEN
       CALL MNH_GET_GT3D_N0(KINDEX)
       IF (KJ .EQ. IJU) THEN 
          PTAB => GT3D(:,:,1,KINDEX)
       ELSE IF (KJ .EQ. 0)  THEN
          PTAB => GT2D_OSIZE
       ELSE
          call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_GT2D', ' Size mismatsh ' )
       END IF
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_GT2D', ' Size mismatsh ' )
    END IF
       
  END FUNCTION MNH_ALLOCATE_GT2D

  !------------ End Logical Pool

    !-------- Real 1D Pool Managment
  
    SUBROUTINE MNH_GET_ZT1D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF (NT1D_TOP_R == JPMAX_T1D_R ) THEN
      WRITE( *, '( " MNH_GET_ZT1D_N0: NT1D_TOP_R too big (increaze JPMAX_T1D_R) , NT1D_TOP_R=",I4 )' ) NT1D_TOP_R
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_ZT1D_N0', 'NT1D_TOP_R too big (increaze JPMAX_T1D_R)' )
    ELSE
       NT1D_TOP_R = NT1D_TOP_R + 1
       KTEMP = NT1D_POOL_R(NT1D_TOP_R)
       NT1D_POOL_R(NT1D_TOP_R) = - KTEMP
       IF (  NT1D_TOP_R > NT1D_TOP_R_MAX ) THEN
          NT1D_TOP_R_MAX = NT1D_TOP_R
          !WRITE( *, '( " MNH_GET_ZT1D: NT1D_TOP_R_MAX=",I4," KTEMP=",I4 )' ) NT1D_TOP_R_MAX,KTEMP
       END IF
    ENDIF
    !WRITE( *, '( "MNH_GET_ZT1D: reserving ZT1D (",I4,")" )' ) KTEMP

  END SUBROUTINE MNH_GET_ZT1D_N0

  SUBROUTINE MNH_GET_ZT1D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18

                          CALL MNH_GET_ZT1D_N0(KTEMP1)
    IF (PRESENT(KTEMP2))  CALL MNH_GET_ZT1D_N0(KTEMP2)
    IF (PRESENT(KTEMP3))  CALL MNH_GET_ZT1D_N0(KTEMP3)
    IF (PRESENT(KTEMP4))  CALL MNH_GET_ZT1D_N0(KTEMP4)
    IF (PRESENT(KTEMP5))  CALL MNH_GET_ZT1D_N0(KTEMP5)
    IF (PRESENT(KTEMP6))  CALL MNH_GET_ZT1D_N0(KTEMP6)
    IF (PRESENT(KTEMP7))  CALL MNH_GET_ZT1D_N0(KTEMP7)
    IF (PRESENT(KTEMP8))  CALL MNH_GET_ZT1D_N0(KTEMP8)
    IF (PRESENT(KTEMP9))  CALL MNH_GET_ZT1D_N0(KTEMP9)
    IF (PRESENT(KTEMP10)) CALL MNH_GET_ZT1D_N0(KTEMP10)
    IF (PRESENT(KTEMP11)) CALL MNH_GET_ZT1D_N0(KTEMP11)
    IF (PRESENT(KTEMP12)) CALL MNH_GET_ZT1D_N0(KTEMP12)
    IF (PRESENT(KTEMP13)) CALL MNH_GET_ZT1D_N0(KTEMP13)
    IF (PRESENT(KTEMP14)) CALL MNH_GET_ZT1D_N0(KTEMP14)
    IF (PRESENT(KTEMP15)) CALL MNH_GET_ZT1D_N0(KTEMP15)
    IF (PRESENT(KTEMP16)) CALL MNH_GET_ZT1D_N0(KTEMP16)
    IF (PRESENT(KTEMP17)) CALL MNH_GET_ZT1D_N0(KTEMP17)
    IF (PRESENT(KTEMP18)) CALL MNH_GET_ZT1D_N0(KTEMP18)

  END SUBROUTINE MNH_GET_ZT1D

    SUBROUTINE MNH_REL_ZT1D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF ( ( NT1D_TOP_R > JPMAX_T1D_R ) .OR. ( NT1D_TOP_R < 1 ) ) THEN
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_ZT1D_N0', 'invalid value for NT1D_TOP_R' )
    ELSE
      NT1D_POOL_R(KTEMP) = KTEMP
      IF (KTEMP == NT1D_TOP_R) THEN
        NT1D_TOP_R = NT1D_TOP_R - 1
     ELSE
        WRITE( *, '( "MNH_REL_ZT1D_N0: invalid value for KTEMP <> NT1D_TOP_R  (",2I8,")" )' ) KTEMP, NT1D_TOP_R
        FLUSH(6)
        call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_ZT1D_N0', 'invalid value for KTEMP <> NT1D_TOP_R' )
      END IF
    ENDIF
    !WRITE( *, '( "MNH_REL_ZT1D_N0: releasing ZT1D (",I4,")" )' ) KTEMP

  END SUBROUTINE MNH_REL_ZT1D_N0

  SUBROUTINE MNH_REL_ZT1D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18, &
                          KTEMP19,KTEMP20)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18
    INTEGER,OPTIONAL :: KTEMP19,KTEMP20

    IF (PRESENT(KTEMP20)) CALL MNH_REL_ZT1D_N0(KTEMP20)
    IF (PRESENT(KTEMP19)) CALL MNH_REL_ZT1D_N0(KTEMP19)
    IF (PRESENT(KTEMP18)) CALL MNH_REL_ZT1D_N0(KTEMP18)
    IF (PRESENT(KTEMP17)) CALL MNH_REL_ZT1D_N0(KTEMP17)
    IF (PRESENT(KTEMP16)) CALL MNH_REL_ZT1D_N0(KTEMP16)
    IF (PRESENT(KTEMP15)) CALL MNH_REL_ZT1D_N0(KTEMP15)
    IF (PRESENT(KTEMP14)) CALL MNH_REL_ZT1D_N0(KTEMP14)
    IF (PRESENT(KTEMP13)) CALL MNH_REL_ZT1D_N0(KTEMP13)
    IF (PRESENT(KTEMP12)) CALL MNH_REL_ZT1D_N0(KTEMP12)
    IF (PRESENT(KTEMP11)) CALL MNH_REL_ZT1D_N0(KTEMP11)
    IF (PRESENT(KTEMP10)) CALL MNH_REL_ZT1D_N0(KTEMP10)
    IF (PRESENT(KTEMP9))  CALL MNH_REL_ZT1D_N0(KTEMP9)
    IF (PRESENT(KTEMP8))  CALL MNH_REL_ZT1D_N0(KTEMP8)
    IF (PRESENT(KTEMP7))  CALL MNH_REL_ZT1D_N0(KTEMP7)
    IF (PRESENT(KTEMP6))  CALL MNH_REL_ZT1D_N0(KTEMP6)
    IF (PRESENT(KTEMP5))  CALL MNH_REL_ZT1D_N0(KTEMP5)
    IF (PRESENT(KTEMP4))  CALL MNH_REL_ZT1D_N0(KTEMP4)
    IF (PRESENT(KTEMP3))  CALL MNH_REL_ZT1D_N0(KTEMP3)
    IF (PRESENT(KTEMP2))  CALL MNH_REL_ZT1D_N0(KTEMP2)
    CALL MNH_REL_ZT1D_N0(KTEMP1)
    
  END SUBROUTINE MNH_REL_ZT1D

  FUNCTION MNH_ALLOCATE_ZT1D(PTAB,KI) RESULT (KINDEX)
    
    REAL, POINTER, CONTIGUOUS , DIMENSION(:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI
    INTEGER                 :: KINDEX

    !local

    IF (KI .EQ. IIJKU) THEN 
       CALL MNH_GET_ZT1D_N0(KINDEX)
       PTAB => ZT1D(:,KINDEX)
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_ZT1D', ' Size mismatsh ' )
    END IF
    !WRITE( *, '( "MNH_ALLOCATE_ZT1D: KI=(",I4,") , KINDEX=(",I4,")" )' ) KI,KINDEX
    !FLUSH(6) 
       
  END FUNCTION MNH_ALLOCATE_ZT1D

  FUNCTION MNH_ALLOCATE_ZT1DP(PTAB,KI) RESULT (KINDEX)
    
    REAL, POINTER, CONTIGUOUS , DIMENSION(:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI
    INTEGER                 :: KINDEX

    !local

    IF (KI .LE. IIJKU) THEN 
       CALL MNH_GET_ZT1D_N0(KINDEX)
       IF (KI .NE. 0) THEN
          PTAB(1:KI) => ZT1D(:,KINDEX)
       ELSE
          PTAB => ZT1D_OSIZE
       END IF
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_ZT1DP', ' Size mismatsh ' )
    END IF
    !WRITE( *, '( "MNH_ALLOCATE_ZT1DP: KI=(",I9,") , KINDEX=(",I4,")" )' ) KI,KINDEX
    
  END FUNCTION MNH_ALLOCATE_ZT1DP

  ! End Real 1D management
  
  !-------- Integer 1D Pool Managment
  
    SUBROUTINE MNH_GET_IT1D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF (NT1D_TOP_I == JPMAX_T1D_I ) THEN
      WRITE( *, '( " MNH_GET_IT1D_N0: NT1D_TOP_I too big (increaze JPMAX_T1D_I) , NT1D_TOP_I=",I4 )' ) NT1D_TOP_I
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_IT1D_N0', 'NT1D_TOP_I too big (increaze JPMAX_T1D_I)' )
    ELSE
       NT1D_TOP_I = NT1D_TOP_I + 1
       KTEMP = NT1D_POOL_I(NT1D_TOP_I)
       NT1D_POOL_I(NT1D_TOP_I) = - KTEMP
       IF (  NT1D_TOP_I > NT1D_TOP_I_MAX ) THEN
          NT1D_TOP_I_MAX = NT1D_TOP_I
          !WRITE( *, '( " MNH_GET_IT1D: NT1D_TOP_I_MAX=",I4," KTEMP=",I4 )' ) NT1D_TOP_I_MAX,KTEMP
       END IF
    ENDIF
    !WRITE( *, '( "MNH_GET_IT1D: reserving IT1D (",I4,")" )' ) KTEMP

  END SUBROUTINE MNH_GET_IT1D_N0

  SUBROUTINE MNH_GET_IT1D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18

                          CALL MNH_GET_IT1D_N0(KTEMP1)
    IF (PRESENT(KTEMP2))  CALL MNH_GET_IT1D_N0(KTEMP2)
    IF (PRESENT(KTEMP3))  CALL MNH_GET_IT1D_N0(KTEMP3)
    IF (PRESENT(KTEMP4))  CALL MNH_GET_IT1D_N0(KTEMP4)
    IF (PRESENT(KTEMP5))  CALL MNH_GET_IT1D_N0(KTEMP5)
    IF (PRESENT(KTEMP6))  CALL MNH_GET_IT1D_N0(KTEMP6)
    IF (PRESENT(KTEMP7))  CALL MNH_GET_IT1D_N0(KTEMP7)
    IF (PRESENT(KTEMP8))  CALL MNH_GET_IT1D_N0(KTEMP8)
    IF (PRESENT(KTEMP9))  CALL MNH_GET_IT1D_N0(KTEMP9)
    IF (PRESENT(KTEMP10)) CALL MNH_GET_IT1D_N0(KTEMP10)
    IF (PRESENT(KTEMP11)) CALL MNH_GET_IT1D_N0(KTEMP11)
    IF (PRESENT(KTEMP12)) CALL MNH_GET_IT1D_N0(KTEMP12)
    IF (PRESENT(KTEMP13)) CALL MNH_GET_IT1D_N0(KTEMP13)
    IF (PRESENT(KTEMP14)) CALL MNH_GET_IT1D_N0(KTEMP14)
    IF (PRESENT(KTEMP15)) CALL MNH_GET_IT1D_N0(KTEMP15)
    IF (PRESENT(KTEMP16)) CALL MNH_GET_IT1D_N0(KTEMP16)
    IF (PRESENT(KTEMP17)) CALL MNH_GET_IT1D_N0(KTEMP17)
    IF (PRESENT(KTEMP18)) CALL MNH_GET_IT1D_N0(KTEMP18)

  END SUBROUTINE MNH_GET_IT1D

    SUBROUTINE MNH_REL_IT1D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF ( ( NT1D_TOP_I > JPMAX_T1D_I ) .OR. ( NT1D_TOP_I < 1 ) ) THEN
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_IT1D_N0', 'invalid value for NT1D_TOP_I' )
    ELSE
      NT1D_POOL_I(KTEMP) = KTEMP
      IF (KTEMP == NT1D_TOP_I) THEN
        NT1D_TOP_I = NT1D_TOP_I - 1
     ELSE
        WRITE( *, '( "MNH_REL_IT1D_N0: invalid value for KTEMP <> NT1D_TOP_I  (",2I8,")" )' ) KTEMP, NT1D_TOP_I
        FLUSH(6)
        call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_IT1D_N0', 'invalid value for KTEMP <> NT1D_TOP_R' )
      END IF
    ENDIF
    !WRITE( *, '( "MNH_REL_IT1D_N0: releasing IT1D (",I4,")" )' ) KTEMP

  END SUBROUTINE MNH_REL_IT1D_N0

  SUBROUTINE MNH_REL_IT1D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18, &
                          KTEMP19,KTEMP20)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18
    INTEGER,OPTIONAL :: KTEMP19,KTEMP20

    IF (PRESENT(KTEMP20)) CALL MNH_REL_IT1D_N0(KTEMP20)
    IF (PRESENT(KTEMP19)) CALL MNH_REL_IT1D_N0(KTEMP19)
    IF (PRESENT(KTEMP18)) CALL MNH_REL_IT1D_N0(KTEMP18)
    IF (PRESENT(KTEMP17)) CALL MNH_REL_IT1D_N0(KTEMP17)
    IF (PRESENT(KTEMP16)) CALL MNH_REL_IT1D_N0(KTEMP16)
    IF (PRESENT(KTEMP15)) CALL MNH_REL_IT1D_N0(KTEMP15)
    IF (PRESENT(KTEMP14)) CALL MNH_REL_IT1D_N0(KTEMP14)
    IF (PRESENT(KTEMP13)) CALL MNH_REL_IT1D_N0(KTEMP13)
    IF (PRESENT(KTEMP12)) CALL MNH_REL_IT1D_N0(KTEMP12)
    IF (PRESENT(KTEMP11)) CALL MNH_REL_IT1D_N0(KTEMP11)
    IF (PRESENT(KTEMP10)) CALL MNH_REL_IT1D_N0(KTEMP10)
    IF (PRESENT(KTEMP9))  CALL MNH_REL_IT1D_N0(KTEMP9)
    IF (PRESENT(KTEMP8))  CALL MNH_REL_IT1D_N0(KTEMP8)
    IF (PRESENT(KTEMP7))  CALL MNH_REL_IT1D_N0(KTEMP7)
    IF (PRESENT(KTEMP6))  CALL MNH_REL_IT1D_N0(KTEMP6)
    IF (PRESENT(KTEMP5))  CALL MNH_REL_IT1D_N0(KTEMP5)
    IF (PRESENT(KTEMP4))  CALL MNH_REL_IT1D_N0(KTEMP4)
    IF (PRESENT(KTEMP3))  CALL MNH_REL_IT1D_N0(KTEMP3)
    IF (PRESENT(KTEMP2))  CALL MNH_REL_IT1D_N0(KTEMP2)
    CALL MNH_REL_IT1D_N0(KTEMP1)
    
  END SUBROUTINE MNH_REL_IT1D

  FUNCTION MNH_ALLOCATE_IT1D(PTAB,KI) RESULT (KINDEX)
    
    INTEGER, POINTER, CONTIGUOUS , DIMENSION(:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI
    INTEGER                 :: KINDEX

    !local

    IF (KI .EQ. IIJKU) THEN 
       CALL MNH_GET_IT1D_N0(KINDEX)
       PTAB => IT1D(:,KINDEX)
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_IT1D', ' Size mismatsh ' )
    END IF
    !WRITE( *, '( "MNH_ALLOCATE_IT1D: KI=(",I4,") , KINDEX=(",I4,")" )' ) KI,KINDEX
    !FLUSH(6) 
       
  END FUNCTION MNH_ALLOCATE_IT1D

  FUNCTION MNH_ALLOCATE_IT1DP(PTAB,KI) RESULT (KINDEX)
    
    INTEGER, POINTER, CONTIGUOUS , DIMENSION(:), INTENT(INOUT) :: PTAB
    INTEGER , INTENT(IN)    :: KI
    INTEGER                 :: KINDEX

    !local

    IF (KI .LE. IIJKU) THEN 
       CALL MNH_GET_IT1D_N0(KINDEX)
       IF (KI .NE. 0) THEN
          PTAB(1:KI) => IT1D(:,KINDEX)
       ELSE
          PTAB => IT1D_OSIZE
       END IF
    ELSE
       call Print_msg( NVERB_ERROR, 'GEN', 'MNH_ALLOCATE_IT1DP', ' Size mismatsh ' )
    END IF
    !WRITE( *, '( "MNH_ALLOCATE_IT1DP: KI=(",I9,") , KINDEX=(",I4,")" )' ) KI,KINDEX
    
  END FUNCTION MNH_ALLOCATE_IT1DP

  ! End Integer 1D management

  SUBROUTINE MNH_CHECK_IN_ZT3D(HSUB)
    IMPLICIT NONE

    CHARACTER(LEN=*) :: HSUB

    !print*,"MNH_CHECK_IN_ZT3D => " , NT3D_TOP_CURRENT_INDEX+1 ,  HSUB
    NT3D_TOP_CURRENT_INDEX = NT3D_TOP_CURRENT_INDEX + 1
    NT3D_TOP_CURRENT(NT3D_TOP_CURRENT_INDEX) = NT3D_TOP
  END SUBROUTINE MNH_CHECK_IN_ZT3D

  SUBROUTINE MNH_CHECK_OUT_ZT3D(HSUB)
    IMPLICIT NONE

    CHARACTER(LEN=*) :: HSUB
    
    !print*,"MNH_CHECK_OUT_ZT3D <= " ,  NT3D_TOP_CURRENT_INDEX , HSUB
    IF ( NT3D_TOP_CURRENT(NT3D_TOP_CURRENT_INDEX) .NE.  NT3D_TOP ) THEN
        WRITE( *, '( "MNH_CHECK_OUT_ZT3D : NT3D_TOP_CURRENT .NE. NT3D_TOP (",2I8,")" )' )  NT3D_TOP_CURRENT ,  NT3D_TOP
        call Print_msg( NVERB_ERROR, 'GEN', 'MNH_CHECK_OUT_ZT3D', ' CHECK IN/OUT MISTASK ' )
     ELSE
        NT3D_TOP_CURRENT_INDEX = NT3D_TOP_CURRENT_INDEX - 1
    END IF 
  END SUBROUTINE MNH_CHECK_OUT_ZT3D

END MODULE MODE_MNH_ZWORK
#endif
